from __future__ import division, print_function, absolute_import, unicode_literals

from unicodedata import east_asian_width
import six

from mog_commons.collection import distinct
from mog_commons.types import *

__all__ = [
    'is_unicode',
    'is_strlike',
    'unicode_width',
    'to_unicode',
    'to_str',
    'to_bytes',
    'edge_just',
    'unicode_ljust',
    'unicode_rjust',
    'unicode_right',
    'unicode_left',
    'unicode_decode',
]

__unicode_width_mapping = {'F': 2, 'H': 1, 'W': 2, 'Na': 1, 'A': 2, 'N': 1}


def is_unicode(s):
    return (six.PY2 and isinstance(s, unicode)) or (six.PY3 and isinstance(s, str))


def is_strlike(s):
    return isinstance(s, (six.string_types, bytes))


@types(s=String)
def unicode_width(s):
    if is_unicode(s):
        return sum(__unicode_width_mapping[east_asian_width(c)] for c in s)

    assert is_strlike(s), 's must be a string, not %s.' % type(s).__name__
    return len(s)


@types(encoding=Option(String), errors=String)
def to_unicode(s, encoding=None, errors='strict'):
    """
    Make unicode string from any value
    :param s:
    :param encoding:
    :param errors:
    :return: unicode
    """
    encoding = encoding or 'utf-8'

    if is_unicode(s):
        return s
    elif is_strlike(s):
        return s.decode(encoding, errors)
    else:
        if six.PY2:
            return str(s).decode(encoding, errors)
        else:
            return str(s)


@types(encoding=Option(String), errors=String)
def to_str(s, encoding=None, errors='strict'):
    """
    Make str from any value
    :param s:
    :param encoding:
    :param errors:
    :return: str (not unicode in Python2, nor bytes in Python3)
    """
    encoding = encoding or 'utf-8'

    if is_strlike(s):
        if six.PY2:
            return s.encode(encoding, errors) if isinstance(s, unicode) else s
        else:
            return s.decode(encoding, errors) if isinstance(s, bytes) else s
    else:
        return str(s)


@types(encoding=Option(String), errors=String)
def to_bytes(s, encoding=None, errors='strict'):
    """Convert string to bytes."""
    encoding = encoding or 'utf-8'

    if is_unicode(s):
        return s.encode(encoding, errors)
    elif is_strlike(s):
        return s
    else:
        if six.PY2:
            return str(s)
        else:
            return str(s).encode(encoding, errors)


@types(left=String, right=String, width=int, fillchar=String, min_padding_length=int)
def edge_just(left, right, width, fillchar=' ', min_padding_length=1):
    """
    :param left:
    :param right:
    :param width:
    :param fillchar:
    :param min_padding_length: minimum padding length
    :return:
    """
    assert unicode_width(fillchar) == 1, 'fillchar must be single-width char'
    padding = fillchar * max(min_padding_length, width - unicode_width(left + right))
    return left + padding + right


@types(s=String, width=int, fillchar=String)
def unicode_ljust(s, width, fillchar=' '):
    """
    Return Unicode string left-justified in a print-length width.
    Padding is done using the specified fill character (default is a space).
    """
    return edge_just(s, '', width, fillchar, 0)


@types(s=String, width=int, fillchar=String)
def unicode_rjust(s, width, fillchar=' '):
    """
    Return Unicode string right-justified in a print-length width.
    Padding is done using the specified fill character (default is a space).
    """
    return edge_just('', s, width, fillchar, 0)


@types(s=String, width=int)
def unicode_left(s, width):
    """Cut unicode string from left to fit a given width."""
    i = 0
    j = 0
    for ch in s:
        j += __unicode_width_mapping[east_asian_width(ch)]
        if width < j:
            break
        i += 1
    return s[:i]


@types(s=String, width=int)
def unicode_right(s, width):
    """Cut unicode string from right to fit a given width."""
    i = len(s)
    j = 0
    for ch in reversed(s):
        j += __unicode_width_mapping[east_asian_width(ch)]
        if width < j:
            break
        i -= 1
    return s[i:]


@types(encoding_list=(String, ListOf(String)))
def unicode_decode(data, encoding_list):
    """
    Decode string data with one or more encodings, trying sequentially
    :param data: bytes: encoded string data
    :param encoding_list: list[string] or string: encoding names
    :return: string: decoded string
    """
    assert encoding_list, 'encodings must not be empty.'

    xs = distinct(encoding_list if isinstance(encoding_list, list) else [encoding_list])
    first_exp = None
    for i, encoding in enumerate(xs):
        try:
            return data.decode(encoding)
        except UnicodeDecodeError as e:
            if i == 0:
                first_exp = e
    raise first_exp



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import sys
import codecs
import subprocess
import locale
import platform
import time

if os.name == 'nt':
    # for Windows
    import msvcrt
else:
    # for Unix/Linux/Mac/CygWin
    import termios
    import tty

from mog_commons.case_class import CaseClass
from mog_commons.string import to_unicode

__all__ = [
    'TerminalHandler',
]

DEFAULT_GETCH_REPEAT_THRESHOLD = 0.3  # in seconds


class TerminalHandler(CaseClass):
    """
    IMPORTANT: When you use this class in POSIX environment, make sure to set signal function for restoring terminal
    attributes. The function `restore_terminal` is for that purpose. See the example below.

    :example:
            import signal

            t = TerminalHandler()
            signal.signal(signal.SIGTERM, t.restore_terminal)

            try:
                (do your work)
            finally:
                t.restore_terminal(None, None)
    """

    def __init__(self, term_type=None, encoding=None,
                 stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr,
                 getch_repeat_threshold=DEFAULT_GETCH_REPEAT_THRESHOLD,
                 keep_input_clean=True, getch_enabled=True):
        CaseClass.__init__(self,
                           ('term_type', term_type or self._detect_term_type()),
                           ('encoding', encoding or self._detect_encoding(stdout)),
                           ('stdin', stdin),
                           ('stdout', stdout),
                           ('stderr', stderr),
                           ('getch_repeat_threshold', getch_repeat_threshold),
                           ('keep_input_clean', keep_input_clean),
                           ('getch_enabled', getch_enabled and self._can_getch_enable(stdin))
                           )
        self.restore_terminal = self._get_restore_function()  # binary function for restoring terminal attributes
        self.last_getch_time = 0.0
        self.last_getch_char = '..'

    @staticmethod
    def _can_getch_enable(stdin):
        if stdin.isatty():
            return os.name == 'nt' or hasattr(stdin, 'fileno')
        return False

    @staticmethod
    def _detect_term_type():
        """
        Detect the type of the terminal.
        """
        if os.name == 'nt':
            if os.environ.get('TERM') == 'xterm':
                # maybe MinTTY
                return 'mintty'
            else:
                return 'nt'
        if platform.system().upper().startswith('CYGWIN'):
            return 'cygwin'
        return 'posix'

    @staticmethod
    def _detect_encoding(stdout):
        """
        Detect the default encoding for the terminal's output.
        :return: string: encoding string
        """
        if hasattr(stdout, 'encoding') and stdout.encoding:
            return stdout.encoding

        if os.environ.get('LANG'):
            encoding = os.environ.get('LANG').split('.')[-1]

            # validate the encoding string
            ret = None
            try:
                ret = codecs.lookup(encoding)
            except LookupError:
                pass
            if ret:
                return encoding

        return locale.getpreferredencoding()

    def _get_restore_function(self):
        """
        Return the binary function for restoring terminal attributes.
        :return: function (signal, frame) => None:
        """
        if os.name == 'nt' or not self.getch_enabled:
            return lambda signal, frame: None

        try:
            fd = self.stdin.fileno()
            initial = termios.tcgetattr(fd)
        except termios.error:
            return lambda signal, frame: None

        return lambda signal, frame: termios.tcsetattr(fd, termios.TCSADRAIN, initial)

    def clear(self):
        """
        Clear the terminal screen.
        """
        if self.stdout.isatty() or self.term_type == 'mintty':
            cmd, shell = {
                'posix': ('clear', False),
                'nt': ('cls', True),
                'cygwin': (['echo', '-en', r'\ec'], False),
                'mintty': (r'echo -en "\ec', False),
            }[self.term_type]
            subprocess.call(cmd, shell=shell, stdin=self.stdin, stdout=self.stdout, stderr=self.stderr)

    def clear_input_buffer(self):
        """
        Clear the input buffer.
        """
        if self.stdin.isatty():
            if os.name == 'nt':
                while msvcrt.kbhit():
                    msvcrt.getch()
            else:
                try:
                    self.stdin.seek(0, 2)  # may fail in some unseekable file object
                except IOError:
                    pass

    def getch(self):
        """
        Read one character from stdin.

        If stdin is not a tty or set `getch_enabled`=False, read input as one line.
        :return: unicode:
        """
        ch = self._get_one_char()
        if self.keep_input_clean:
            self.clear_input_buffer()

        try:
            # accept only unicode characters (for Python 2)
            uch = to_unicode(ch, 'ascii')
        except UnicodeError:
            return ''

        return uch if self._check_key_repeat(uch) else ''

    def _get_one_char(self):
        if not self.getch_enabled:
            return self.gets()[:1]
        elif os.name == 'nt':  # Windows
            return msvcrt.getch()
        else:  # POSIX
            try:
                tty.setraw(self.stdin.fileno())
                return self.stdin.read(1)
            finally:
                self.restore_terminal(None, None)

    def _check_key_repeat(self, ch):
        if self.getch_repeat_threshold <= 0.0:
            return True

        t = time.time()
        if ch == self.last_getch_char and t < self.last_getch_time + self.getch_repeat_threshold:
            self.last_getch_time = t
            return False

        self.last_getch_time = t
        self.last_getch_char = ch
        return True

    def gets(self):
        """
        Read line from stdin.

        The trailing newline will be omitted.
        :return: string:
        """
        ret = self.stdin.readline()
        if ret == '':
            raise EOFError  # To break out of EOF loop
        return ret.rstrip('\n')



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import six
import io
import tempfile
from contextlib import contextmanager
import jinja2

# unittest
if sys.version_info < (2, 7):
    import unittest2 as base_unittest
else:
    import unittest as base_unittest

from mog_commons.string import to_bytes, to_str

__all__ = [
    'FakeInput',
    'FakeBytesInput',
    'TestCase',
]


class StringBuffer(object):
    """
    Replace for six.StringIO

    We don't use StringIO because there are many differences between PY2 and PY3.
    """

    def __init__(self, init_buffer=None):
        self._buffer = init_buffer or b''

    def write(self, s, encoding='utf-8', errors='strict'):
        self._buffer += to_bytes(s, encoding, errors)

    def writelines(self, lines, encoding='utf-8', errors='strict'):
        self._buffer += b''.join(to_bytes(s, encoding, errors) for s in lines)

    def flush(self):
        """do nothing"""

    def getvalue(self, encoding='utf-8', errors='strict'):
        return self._buffer.decode(encoding, errors)


class FakeInput(six.StringIO):
    """Fake input object"""

    def __init__(self, buff=None):
        six.StringIO.__init__(self, buff or '')

    def fileno(self):
        return 0

    def isatty(self):
        return True


class FakeBytesInput(six.BytesIO):
    """Fake bytes input object"""

    def __init__(self, buff=None):
        six.BytesIO.__init__(self, buff or b'')

    def fileno(self):
        return 0

    def isatty(self):
        return True


class TestCase(base_unittest.TestCase):
    def assertRaisesRegexp(self, expected_exception, expected_regexp, callable_obj=None, *args, **kwargs):
        """
        Accept difference of the function name between PY2 and PY3.

        We don't use built-in assertRaisesRegexp because it is unicode-unsafe.
        """
        encoding = 'utf-8'
        with self.assertRaises(expected_exception) as cm:
            callable_obj(*args, **kwargs)
        if six.PY2:
            try:
                msg = to_str(cm.exception, encoding)
            except UnicodeEncodeError:
                # avoid to use cm.exception.message
                msg = cm.exception.args[0]
            self.assertRegexpMatches(msg, expected_regexp)
        else:
            self.assertRegex(to_str(cm.exception, encoding), expected_regexp)

    def assertRaisesMessage(self, expected_exception, expected_message, callable_obj=None, *args, **kwargs):
        """
        Assert the expected exception is raised and its message is equal to expected.

        :param expected_exception: class:
        :param expected_message: string:
        :param callable_obj: function:
        :param args: args
        :param kwargs: kwargs
        """
        encoding = 'utf-8'
        with self.assertRaises(expected_exception) as cm:
            callable_obj(*args, **kwargs)
        try:
            msg = to_str(cm.exception, encoding)
        except UnicodeEncodeError:
            # avoid to use cm.exception.message
            msg = cm.exception.args[0]
        self.assertEqual(msg, expected_message)

    @contextmanager
    def withOutput(self):
        """
        Capture and suppress stdout and stderr.
        example:
            with self.withOutput() as (out, err):
                (do main logic)
            (verify out.getvalue() or err.getvalue())
        """
        new_out, new_err = StringBuffer(), StringBuffer()
        old_out, old_err = sys.stdout, sys.stderr

        try:
            sys.stdout, sys.stderr = new_out, new_err
            yield sys.stdout, sys.stderr
        finally:
            sys.stdout, sys.stderr = old_out, old_err

    @contextmanager
    def withAssertOutput(self, expected_stdout, expected_stderr, encoding='utf-8'):
        with self.withOutput() as (out, err):
            yield out, err
        self.assertMultiLineEqual(out.getvalue(encoding), expected_stdout)
        self.assertMultiLineEqual(err.getvalue(encoding), expected_stderr)

    def assertOutput(self, expected_stdout, expected_stderr, function, encoding='utf-8'):
        with self.withAssertOutput(expected_stdout, expected_stderr, encoding) as (out, err):
            function()

    @contextmanager
    def withAssertOutputFile(self, expect_file, variables=None, expect_file_encoding='utf-8', output_encoding='utf-8'):
        """
        Create a temporary file as output and compare with the file content.

        :param expect_file: string: path to the file which contains the expected output
        :param variables: dict: variables for template engine jinja2
        :param expect_file_encoding: string:
        :param output_encoding: string:
        """
        with tempfile.TemporaryFile() as out:
            yield out

            with io.open(expect_file, encoding=expect_file_encoding) as f:
                expect = f.read()
                if variables:
                    expect = jinja2.Template(expect).render(**variables)

            out.seek(0)
            actual = out.read().decode(output_encoding)
            self.assertMultiLineEqual(actual, expect)

    def assertSystemExit(self, expected_code, callable_obj=None, *args, **kwargs):
        """
        Fail unless SystemExit is raised by callableObj when invoked with arguments args and keyword arguments kwargs.
        :param expect_code:
        :param function:
        :return:
        """
        with self.assertRaises(SystemExit) as cm:
            callable_obj(*args, **kwargs)
        actual_code = cm.exception if isinstance(cm.exception, int) else cm.exception.code
        self.assertEqual(actual_code, expected_code)
